/*
 */

#include <linux/version.h>

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(x,y,z) (((x)<<16)+((y)<<8)+(z))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
  #error not suported
#else
  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
    #define KERNEL22
  #else
    #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
      #error not suported
    #else
      #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0)
        #define KERNEL20
      #else
        #error not suported
      #endif
    #endif
  #endif
#endif

#include <linux/module.h>

#ifdef KERNEL22
#include <linux/init.h>
#endif

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <linux/config.h>
#include <linux/pci.h>
#include <linux/miscdevice.h>
#include <linux/ioctl.h>

#ifndef KERNEL22
#include <linux/bios32.h>
#endif

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/system.h>

#ifdef KERNEL22
#include <asm/uaccess.h>
#endif

#ifdef STANDALONE
#include "piix4acpi.h"
#else
#include <linux/piix4acpi.h>
#endif

#ifndef PIIX4ACPI_MINOR
#define PIIX4ACPI_MINOR 160
#endif


static unsigned short apmiobase=0;

static int  piix4acpi_ioctl(   struct inode *, struct file *,
			  unsigned int, unsigned long );
static int  piix4acpi_open(    struct inode *, struct file * );
#ifdef KERNEL22
static int  piix4acpi_release( struct inode *, struct file * );
#else
static void piix4acpi_release( struct inode *, struct file * );
#endif
static int piix4acpi_find(void);



static struct file_operations piix4acpi_fops = {
	NULL,           /* lseek  */
	NULL,           /* read  */
	NULL,           /* write */
	NULL,		/* readdir */
	NULL,		/* select/poll */
	piix4acpi_ioctl,
	NULL,		/* mmap */
	piix4acpi_open,
#ifdef KERNEL22
	NULL,           /* flush */
#endif
	piix4acpi_release
#ifdef KERNEL22
	,NULL,NULL,NULL /* fsync, fasync, check_media_change */
	,NULL,NULL      /* revalidate, lock */
#endif
};

static struct miscdevice piix4acpi_misc = {
	PIIX4ACPI_MINOR, "piix4acpi", &piix4acpi_fops
};



static int piix4acpi_open(struct inode * inode, struct file * file)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
	return 0;
}

#ifdef KERNEL22
static int  piix4acpi_release(struct inode * inode, struct file * file)
#else
static void piix4acpi_release(struct inode * inode, struct file * file)
#endif
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif

#ifdef KERNEL22
	return 0;
#endif
}



static int piix4acpi_ioctl(struct inode *inode, struct file *file,
		      unsigned int cmd, unsigned long arg)
{
static	int pressed_count = 0;
	int retval = 0;
	int p;

#ifdef ACPI_DEBUG
	printk( KERN_INFO "acpi: ioctl, cmd: 0x%x\n", cmd );
#endif

	switch ( cmd ) {
		case PIIX4ACPI_IOCGETSTAT:
			p = inw_p( apmiobase +0 );
			if( p & 0x100 ){ /* check Power Button Status */
				/* reset Power Button Status bit */
				outw_p( 0x100, apmiobase +0 );
				++pressed_count;
			}
#ifdef KERNEL22
			copy_to_user( (void *)arg, &pressed_count,
				      sizeof(int) );
#else
			retval = verify_area(VERIFY_WRITE, 
					     (void *) arg, sizeof(int));
			if (retval){
			    return retval;
			}
			memcpy_tofs( (void *)arg, &pressed_count,
				     sizeof(int) );
#endif
			break;
		default:
			retval = -EINVAL;
	}

	return retval;
}



static int device_ids[]={
	PCI_DEVICE_ID_INTEL_82371AB_3,
	-1 };


static int piix4acpi_find(void)
{
#ifdef CONFIG_PCI
	int i, idx, res;
	int base;
/*
	unsigned char pmreg;
*/
#ifdef KERNEL22
	struct pci_dev *pdev;
#endif
	unsigned char bus, func;


        if (!pcibios_present()){
                return -ENODEV;
	}

	for( i= 0; device_ids[i] >= 0; ++i ){
		printk( KERN_INFO "acpi: searching i=%d id=0x%x\n",
			i, device_ids[i] );
#ifdef KERNEL22
	        pdev = NULL;
#endif
        	for (idx = 0; ; idx++) {
#ifdef KERNEL22
			pdev = pci_find_device( PCI_VENDOR_ID_INTEL,
				device_ids[i], pdev );
			if( pdev == NULL ){
				break;
			}
			func  = pdev->devfn;
			bus   = pdev->bus->number;
#else
                        res = pcibios_find_device(PCI_VENDOR_ID_INTEL, 
					  device_ids[i], idx, &bus, &func);
			if( res != PCIBIOS_SUCCESSFUL ){
				break;
			}
#endif
			printk( KERN_INFO
				"acpi: bus = %d, dev = %d, func = %d\n",
				bus, func/8, func&7 );

			if( (func&0x7) == 3 ){
				/* check Power Management IO Space 
				   has been enabled */
				/*
				res = pcibios_read_config_byte(	bus, func,
					      ACPI_PCICR_PMREGMISC, &pmreg );
				if( res ){
					return -EIO;
				}

				printk( KERN_INFO "acpi: PMREGMISC 0x%x\n",
					pmreg );
				*/
				/* enable Power Management IO Space */
				res = pcibios_write_config_byte( bus, func,
					ACPI_PCICR_PMREGMISC, 0x01 );
				if( res ){
					return -EIO;
				}
				/* get Power Management IO base address */
				res = pcibios_read_config_dword( bus, func,
					ACPI_PCICR_PMBA, &base );
				if( res ){
					return -EIO;
				}
				base &= 0xfffe;
				/* disable Power Button Ebable bit */
				outw_p( 0x0, base+2);
				return( base );
			}
		}
	}
#endif
	return -ENODEV;
}

#ifdef KERNEL22
__initfunc(int piix4acpi_init(void))
#else
int piix4acpi_init(void)
#endif
{
	unsigned int base;

	base = piix4acpi_find();
	if( base < 0 ){
		printk( "acpi: not found\n" );
		return base;
	}

	printk( "acpi: Power Management IO space found at 0x%x\n", base );

	apmiobase = base;

	/* regist misc device */
	misc_register( &piix4acpi_misc );

	/* request IO space */
	request_region( apmiobase, ACPI_APMIOSIZE, "piix4acpi");

	return 0;
}

#ifdef MODULE
int init_module(void)
{
	return piix4acpi_init();
}
#endif

#ifdef MODULE
void cleanup_module(void)
{
	/* deregist misc device */
	misc_deregister( &piix4acpi_misc );

	/* release IO space */
	release_region(  apmiobase, ACPI_APMIOSIZE );
	printk(KERN_INFO "acpi released.\n" );
}
#endif
